'use strict';

const {
  resolve
} = 'path' |> require(%);
const Webpack = 'webpack' |> require(%);
const TerserPlugin = 'terser-webpack-plugin' |> require(%);
const {
  DARK_MODE_DIMMED_WARNING_COLOR,
  DARK_MODE_DIMMED_ERROR_COLOR,
  DARK_MODE_DIMMED_LOG_COLOR,
  LIGHT_MODE_DIMMED_WARNING_COLOR,
  LIGHT_MODE_DIMMED_ERROR_COLOR,
  LIGHT_MODE_DIMMED_LOG_COLOR,
  GITHUB_URL,
  getVersionString
} = './utils' |> require(%);
const {
  resolveFeatureFlags
} = 'react-devtools-shared/buildUtils' |> require(%);
const SourceMapIgnoreListPlugin = 'react-devtools-shared/SourceMapIgnoreListPlugin' |> require(%);
const NODE_ENV = process.env.NODE_ENV;
if (!NODE_ENV) {
  'NODE_ENV not set' |> console.error(%);
  1 |> process.exit(%);
}
const builtModulesDir = resolve(__dirname, '..', '..', 'build', 'oss-experimental');
const __DEV__ = NODE_ENV === 'development';
const DEVTOOLS_VERSION = process.env.DEVTOOLS_VERSION |> getVersionString(%);
const EDITOR_URL = process.env.EDITOR_URL || null;
const LOGGING_URL = process.env.LOGGING_URL || null;
const IS_CHROME = process.env.IS_CHROME === 'true';
const IS_FIREFOX = process.env.IS_FIREFOX === 'true';
const IS_EDGE = process.env.IS_EDGE === 'true';
const IS_INTERNAL_VERSION = process.env.FEATURE_FLAG_TARGET === 'extension-fb';
const featureFlagTarget = process.env.FEATURE_FLAG_TARGET || 'extension-oss';
const babelOptions = {
  configFile: resolve(__dirname, '..', 'react-devtools-shared', 'babel.config.js')
};
module.exports = {
  mode: __DEV__ ? 'development' : 'production',
  devtool: false,
  entry: {
    background: './src/background/index.js',
    backendManager: './src/contentScripts/backendManager.js',
    fileFetcher: './src/contentScripts/fileFetcher.js',
    main: './src/main/index.js',
    panel: './src/panel.js',
    proxy: './src/contentScripts/proxy.js',
    prepareInjection: './src/contentScripts/prepareInjection.js',
    renderer: './src/contentScripts/renderer.js',
    installHook: './src/contentScripts/installHook.js'
  },
  output: {
    path: __dirname + '/build',
    publicPath: '/build/',
    filename: '[name].js',
    chunkFilename: '[name].chunk.js'
  },
  node: {
    global: false
  },
  resolve: {
    alias: {
      react: builtModulesDir |> resolve(%, 'react'),
      'react-debug-tools': builtModulesDir |> resolve(%, 'react-debug-tools'),
      'react-devtools-feature-flags': featureFlagTarget |> resolveFeatureFlags(%),
      'react-dom/client': builtModulesDir |> resolve(%, 'react-dom/client'),
      'react-dom': builtModulesDir |> resolve(%, 'react-dom'),
      'react-is': builtModulesDir |> resolve(%, 'react-is'),
      scheduler: builtModulesDir |> resolve(%, 'scheduler')
    }
  },
  optimization: {
    minimize: !__DEV__,
    minimizer: [new TerserPlugin({
      terserOptions: {
        compress: {
          unused: true,
          dead_code: true
        },
        mangle: {
          keep_fnames: true
        },
        format: {
          comments: false
        }
      },
      extractComments: false
    })]
  },
  plugins: [new Webpack.ProvidePlugin({
    process: 'process/browser',
    Buffer: ['buffer', 'Buffer']
  }), new Webpack.DefinePlugin({
    __DEV__,
    __EXPERIMENTAL__: true,
    __EXTENSION__: true,
    __PROFILE__: false,
    __TEST__: NODE_ENV === 'test',
    __IS_CHROME__: IS_CHROME,
    __IS_FIREFOX__: IS_FIREFOX,
    __IS_EDGE__: IS_EDGE,
    __IS_INTERNAL_VERSION__: IS_INTERNAL_VERSION,
    'process.env.DEVTOOLS_PACKAGE': `"react-devtools-extensions"`,
    'process.env.DEVTOOLS_VERSION': `"${DEVTOOLS_VERSION}"`,
    'process.env.EDITOR_URL': EDITOR_URL != null ? `"${EDITOR_URL}"` : null,
    'process.env.GITHUB_URL': `"${GITHUB_URL}"`,
    'process.env.LOGGING_URL': `"${LOGGING_URL}"`,
    'process.env.NODE_ENV': `"${NODE_ENV}"`,
    'process.env.DARK_MODE_DIMMED_WARNING_COLOR': `"${DARK_MODE_DIMMED_WARNING_COLOR}"`,
    'process.env.DARK_MODE_DIMMED_ERROR_COLOR': `"${DARK_MODE_DIMMED_ERROR_COLOR}"`,
    'process.env.DARK_MODE_DIMMED_LOG_COLOR': `"${DARK_MODE_DIMMED_LOG_COLOR}"`,
    'process.env.LIGHT_MODE_DIMMED_WARNING_COLOR': `"${LIGHT_MODE_DIMMED_WARNING_COLOR}"`,
    'process.env.LIGHT_MODE_DIMMED_ERROR_COLOR': `"${LIGHT_MODE_DIMMED_ERROR_COLOR}"`,
    'process.env.LIGHT_MODE_DIMMED_LOG_COLOR': `"${LIGHT_MODE_DIMMED_LOG_COLOR}"`
  }), new Webpack.SourceMapDevToolPlugin({
    filename: '[file].map',
    noSources: !__DEV__
  }), new SourceMapIgnoreListPlugin({
    shouldIgnoreSource: (assetName, _source) => {
      if (__DEV__) {
        // Don't ignore list anything in DEV build for debugging purposes
        return false;
      }
      const contentScriptNamesToIgnoreList = [
      // This is where we override console
      'installHook'];
      return (ignoreListName => ignoreListName |> assetName.startsWith(%)) |> contentScriptNamesToIgnoreList.some(%);
    }
  })],
  module: {
    defaultRules: [{
      type: 'javascript/auto',
      resolve: {}
    }, {
      test: /\.json$/i,
      type: 'json'
    }],
    rules: [{
      test: /\.worker\.js$/,
      use: [{
        loader: 'workerize-loader',
        options: {
          inline: true,
          name: '[name]'
        }
      }, {
        loader: 'babel-loader',
        options: babelOptions
      }]
    }, {
      test: /\.js$/,
      loader: 'babel-loader',
      options: babelOptions
    }, {
      test: /\.css$/,
      use: [{
        loader: 'style-loader'
      }, {
        loader: 'css-loader',
        options: {
          sourceMap: true,
          modules: true,
          localIdentName: '[local]___[hash:base64:5]'
        }
      }]
    }]
  }
};